(eval-when (:compile-toplevel :load-toplevel :execute)
  (require :asdf)
  (require :gsll)
  (require :iterate)
  (use-package :iterate))
(declaim (optimize (speed 3) (safety 0) (debug 0)))


(defun usable-but-slow-force-function (dim)
  "Given an integer dim, this constructs a function that, when supplied with a 
   N-dimensional vector Z and some output vector (-> pointer?), yields the
   corresponding forces"
  (declare (fixnum dim))
  (let ((temp-values (make-array 2 :element-type 'double-float :initial-element 0.0d0)))
    (lambda (zvector output)
	(do ((i 0 (1+ i))) ((= i dim)) (declare (fixnum i))
	  (setf (aref temp-values 0) 0.0d0)
	  (do ((m 0 (1+ m))) ((> m i)) (declare (fixnum m))
	    (do ((n i (1+ n))) ((= n dim)) (declare (fixnum n))
	      (setf (aref temp-values 1) 0.0d0)
	      (do ((k m (1+ k))) ((> k n)) (declare (fixnum k))
		(incf (aref temp-values 1) (the double-float (grid:gref zvector k))))
	      (incf (aref temp-values 0) (expt (aref temp-values 1) -2))))
	  (setf (the double-float (grid:gref output i)) 
		(- (the double-float (grid:gref zvector i))
		   (aref temp-values 0)))))))

(defun unusable-but-fast-force-function (dim)
  "Given an integer dim, this constructs a function that, when supplied with a 
   N-dimensional vector Z and some output vector (-> pointer?), yields the
   corresponding forces"
  (declare (fixnum dim))
  (let ((cl-output (make-array dim :element-type 'double-float :initial-element 0.0d0))
	(cl-zv (make-array dim :element-type 'double-float :initial-element 0.0d0))
	(temp-values (make-array 2 :element-type 'double-float :initial-element 0.0d0)))
    (declare ((simple-array double-float) cl-output cl-zv))
    (lambda (zvector output)
	(setf cl-output (the (simple-array double-float) (gsl:cl-array output))
	      cl-zv (the (simple-array double-float) (gsl:cl-array zvector))
	 )
	(do ((i 0 (1+ i))) ((= i dim)) (declare (fixnum i))
	  (setf (aref temp-values 0) 0.0d0)
	  (do ((m 0 (1+ m))) ((> m i)) (declare (fixnum m))
	    (do ((n i (1+ n))) ((= n dim)) (declare (fixnum n))
	      (setf (aref temp-values 1) 0.0d0)
	      (do ((k m (1+ k))) ((> k n)) (declare (fixnum k))
		(incf (aref temp-values 1) (aref cl-zv k)))
	      (incf (aref temp-values 0) (expt (aref temp-values 1) -2))))
	  (setf (aref cl-output i) (- (aref cl-zv i) (aref temp-values 0)))))))

(defun force-df (dim)
  "Given an integer dim, this returns a function that operates on one dim-dimensional
   vector v and one dim x dim-dimensional matrix M, storing the Jacobian corresponding to
   force function at the point v in M."
  (declare (fixnum dim))
  (let ((temp-values (make-array 2 :element-type 'double-float
				 :initial-element 0.0d0)))
    (lambda (zvector jacobian)
	(do ((i 0 (1+ i))) ((= i dim))
	  (declare (fixnum i))
	  (do ((j 0 (1+ j))) ((= j dim))
	    (declare (fixnum j))
	    (setf (aref temp-values 0) (if (= i j) 1.0d0 0.0d0))
	    (do ((n (max i j) (1+ n))) ((= n dim))
	      (declare (fixnum n))
	      (do ((m (min i j) (1- m))) ((= m -1))
		(declare (fixnum m))
		(setf (aref temp-values 1) 0.0d0)
		(do ((k m (1+ k))) ((> k n))
		  (declare (fixnum k))
		  (incf (aref temp-values 1) (grid:gref zvector k)))
		(incf (aref temp-values 0) (* 2d0 (expt (aref temp-values 1) -3)))))
	    (setf (grid:gref jacobian i j) (aref temp-values 0)))))))

(defun force-fdf (fn derivative)
  "Given a force-function and a corresponding derivative, this constructs a suitable
   fdf function"
  (declare (function fn derivative))
  (lambda (zvector output jacobian)
    (funcall fn zvector output)
    (funcall derivative zvector jacobian)))

(defun solve-system (dim fn)
  (let* ((df (force-df dim))
	 (fdf (force-fdf fn df))
	 (vec (grid:make-foreign-array
	       'double-float
	       :initial-element 1.0d0
	       :dimensions dim))
	 (solver (gsl:make-multi-dimensional-root-solver-fdf
		  gsl:+powells-hybrid+
		  (list fn df fdf)
		  vec nil))) 
    (iter 
      (declare (iterate:declare-variables))
      (repeat (the fixnum 5))
      (for fnval = (gsl:function-value solver))
      (for argval = (gsl:solution solver))
      (until (gsl:multiroot-test-residual solver 1.0d-7))

      (gsl:iterate solver)
      )))

(let* ((dim 9)
       (vec (grid:make-foreign-array 'double-float :initial-element 1 :dimensions dim))
       (output-2 (grid:make-foreign-array 'double-float :dimensions dim))
       (output-1 (grid:make-foreign-array 'double-float :dimensions dim))
       (mat-1 (grid:make-foreign-array 'double-float :dimensions (list dim dim)))
       (mat-2 (grid:make-foreign-array 'double-float :dimensions (list dim dim)
				       :initial-element 1d0))
       (slow-fn (usable-but-slow-force-function dim))
       (quick-fn (unusable-but-fast-force-function dim)))
  (print "using grid:gref")
  (time (funcall slow-fn vec output-1))
  (print output-1)
  (print "using aref")
  (time (funcall quick-fn vec output-2))
  (print output-2)
  (print "using minimizer with grid:gref")
  (solve-system dim slow-fn)
;  (print "using minimizer with aref")
;  (solve-system dim quick-fn)
  )

